#include <windows.h>
#include <zlib.h>

#pragma pack(push,1)
struct LRFHeader1
{
  //starts at LRF offset 0
  /*00*/ WCHAR signature[4]; //LRF\x00
  /*08*/ WORD version; //version?
  /*0A*/ WORD xorKey;
  /*0C*/ DWORD rootObjectId;
  /*10*/ QWORD nObjects;
  /*18*/ QWORD offObjectTable;
  /*20 - size of header 1 */ 

  void Dump(FILE* out);

};

const wchar_t LRFSignature[4]=L"LRF";

struct LRFHeader2
{
  //starts at LRF offset 20
  /*20*/ DWORD d1_20;
  /*24*/ BYTE  b1_24;
  /*25*/ BYTE  r1_25; //padding
  /*26*/ WORD  w1_26;
  /*28*/ WORD  r2_28; //padding?
  /*2A*/ WORD  w2_2A;
  /*2C*/ WORD  w3_2C;
  /*2E*/ BYTE  b2_2E;
  /*2F*/ BYTE  r3_2F; //padding?
  /*30*/ DWORD signature[5];//20 bytes of something
  /*44*/ DWORD dwTocID;
  /*48*/ DWORD dwTocOffset;
  /*4C*/ WORD  XMLSize; //xml info compressed size
  /*2E size of old version of the header 2 */
//next two only walid if version>=800
  /*4E*/ WORD  w5_4E;
  /*50*/ DWORD GifSize; //gif thumbnail data size
  /*34 size of the header 2 */

  void Dump(FILE* out);
};

#define H2OFF     0x20
#define H2SIZEOLD 0x2E
#define H2SIZE    0x34

struct LRFObjectRecord
{
  DWORD id;      //object ID, must be unique
  DWORD offset;  //LRF offset to object data
  DWORD size;    //size of the object data
  DWORD reserved;//NULL in the file
};

#pragma pack(pop)

class Reader;
class LRFTag;
class LRFFile;

class LRFObject
{
public:
  LRFObject(DWORD id, DWORD offset, DWORD size, LRFFile* file):
      m_dataSize(size), m_id(id), m_offset(offset),
      m_StreamDataOffset(0), m_StreamDataSize(0), m_pFile(file)
  {};
  ~LRFObject()
  {
     Clear();
  }
  virtual bool ReadTags();
  virtual void Dump(FILE* out);
  const std::wstring* name() {return &m_objectName;};

private:
  void Clear();
  bool DumpStream(FILE* out, const std::wstring& filename);
  bool DumpStreamTags(FILE* out, BYTE* buf, DWORD bufLen);
  bool DumpTOC(FILE* out, BYTE* buf, DWORD bufLen);

  DWORD m_dataSize;    //size of the data
  DWORD m_id;          //object ID, must be unique
  DWORD m_offset;      //LRF file offset to the object data
  DWORD m_StreamDataOffset;    //offset to stream data (after F505)
  DWORD m_StreamDataSize;      //size of stream data (set by F504) 
  WORD  m_StreamFlags;         //stream flags(F554): 0x100 = packed, 0x200 = xored, 0x8000 = encrypted
  WORD  m_ObjectType;
  std::vector<LRFTag*> m_tags;
  std::wstring m_objectName;
  LRFFile* m_pFile;
};

typedef std::map<DWORD,LRFObject*> mapDWtoObj;

class Reader
{
public:
  virtual bool Seek(QWORD offset)=0;
  virtual QWORD Pos()=0;
  virtual bool Read(void *buf, QWORD len)=0;
};

class MemReader: public Reader
{
public:
  MemReader(): m_buf(0), m_size(0), m_pos(0) {};
  void SetBuf(BYTE* buf, DWORD size)
  {
    m_buf  = buf;
    m_size = size;
    m_pos  = 0;
  }
  bool Seek(QWORD offset) 
  {
    if (offset>m_size)
      return false;
    else
      m_pos = offset;
    return true;
  };
  QWORD Pos() {return m_pos;};
  bool Read(void *buf, QWORD len)
  {
    if (m_pos+len>m_size)
      return false;
    memcpy(buf, m_buf+m_pos, len);
    m_pos+=len;
    return true;
  }
private:
  BYTE *m_buf;
  DWORD m_size;
  DWORD m_pos;
};

class Descrambler: public Reader
{
public:
  Descrambler(Reader* src, WORD key, DWORD maxlen): m_pSource(src), m_maxLen(maxlen)
  {
    m_xorKey = key & 0xFF;
    m_xorKey = m_xorKey<<8  | m_xorKey; 
    m_xorKey = m_xorKey<<16 | m_xorKey; //propagate low byte to all four   
  };
  QWORD Pos() {return m_pSource->Pos();};
  bool Seek(QWORD offset) {return m_pSource->Seek(offset);}
  bool Read(void *buf, QWORD len)
  {
    if (!m_pSource->Read(buf, len))
      return false;
    DWORD* p = (DWORD*) buf;
    if (len > m_maxLen)
      len = m_maxLen;
    while (len>=4)
    {
      (*p++)^=m_xorKey;
      len-=4;
    }
    BYTE* p2=(BYTE*)p;
    while (len>0)
    {
      (*p2++)^=(m_xorKey&0xFF);
      len--;
    }
    return true;
  }
private:
  DWORD m_xorKey;
  DWORD m_maxLen;
  Reader* m_pSource;
};

class Unpacker: public Reader
{
public:
  Unpacker(Reader* src, DWORD size): m_pSource(src), m_size(size)
  {
  };
  QWORD Pos() {return m_pSource->Pos();};
  bool Seek(QWORD offset) {return m_pSource->Seek(offset);}
  bool Read(void *buf, QWORD len)
  {
    DWORD l(len);
    std::auto_ptr<BYTE> comp(new BYTE[m_size]);
    if (!m_pSource->Read(comp.get(), m_size))
      return false;
    if (Z_OK != uncompress((unsigned char*)buf, &l, comp.get(), m_size))
      return false;
    return true;
  }
private:
  Reader* m_pSource;
  bool b_initedZlib;
  DWORD m_size;
};

class LRFFile: public Reader
{
public:
  LRFFile(const wchar_t* filename):
    m_pHeader1(NULL), m_pHeader2(NULL), m_pFile(NULL), m_filename(filename), m_offsetInLRX(0)
  {};

  bool Seek(QWORD offset)
  {  
     //ugly but works
     size_t offs = offset+m_offsetInLRX;
     return (!fseek(m_pFile, offs, SEEK_SET));
  }

  QWORD Pos()
  {  
     //ugly but works
     return ftell(m_pFile)-m_offsetInLRX;
  }

  bool Read(void *buf, QWORD len)
  {
    size_t lens = len;
    return (lens==fread(buf, 1, lens, m_pFile));
  }

  ~LRFFile()
  {
    fclose(m_pFile);
    delete m_pHeader1;
    delete m_pHeader2;
  }
  
  bool Open();
  void Dump(FILE *out);

  std::wstring m_LastError;
  bool isValidObject(DWORD objectID) {return (m_ObjectsMap.find(objectID) != m_ObjectsMap.end());};
  const std::wstring* getObjectName(DWORD objectID)
  {
    if (isValidObject(objectID))
      return m_ObjectsMap[objectID]->name();
    else
      return new std::wstring(L"<BAD>");
  }
private:

  void DumpObjects(FILE* out);
  bool ReadObjects();
  void DumpMetaInfo(const std::wstring& filename);

  //fields
  
  LRFHeader1* m_pHeader1;
  LRFHeader2* m_pHeader2;
  FILE* m_pFile;
  std::wstring m_filename;
  //maps id to the corresponding object
  mapDWtoObj m_ObjectsMap;
  DWORD m_offsetInLRX;

  friend class LRFObject;
};

class LRFTag
{
  WORD m_tagID;
  BYTE m_shortData[16]; //for standard tags
  BYTE *m_customData;
  DWORD m_dataSize;
public:
  LRFTag(Reader* file): m_tagID(0xF0FF), m_customData(0), m_dataSize(0), m_pFile(file) {};
  ~LRFTag() {delete m_customData;};
  BYTE *data()
  {
    return (m_dataSize>sizeof(m_shortData))?m_customData:m_shortData;
  }
  WORD id() {return m_tagID;};
  DWORD dataSize() {return m_dataSize;};
  bool Load();
  void Dump(FILE* out, LRFFile* pLRFFile);
  BYTE* ReserveBuffer(DWORD size);
  Reader* m_pFile;
};

typedef bool (*CustomTagParser)(Reader* file, LRFTag* tag);

bool DummyTagParser(Reader* file, LRFTag* tag)
{
  fwprintf(stderr, L"Unknown custom tag: 0x%04X at 0x%08I64X\n", tag->id(), file->Pos());
  return false;
}

bool Tag78Parser(Reader* file, LRFTag* tag)
{
  //<dword>, then <FF16> <w:len> <len string> <w2>
  DWORD d1;
  if (!file->Read(&d1, sizeof(d1)))
    return false;
  WORD w;
  if (!file->Read(&w, sizeof(w)))
    return false;
  if (w!=0xF516) //next tag should be F516
    return false;
  if (!file->Read(&w, sizeof(w))) //length of font name
    return false;
  BYTE* buf = tag->ReserveBuffer(w+10);
  *(DWORD*)buf=d1;
  *(WORD*)(buf+6)=w;
  *(WORD*)(buf+w+8)=0;
  if (!file->Read(buf+8, w))
    return false;
  return file->Read(buf+4, 2);
}

bool Tag0B_5CParser(Reader* file, LRFTag* tag)
{
  //<word> count, then count <dword>s
  WORD count;
  if (!file->Read(&count, sizeof(count)))
    return false;
  DWORD size = sizeof(count)+count*sizeof(DWORD);
  BYTE* buf = tag->ReserveBuffer(size);
  *(WORD*)buf = count;
  return file->Read(buf+sizeof(count), size-sizeof(count));
}

bool TagStringParser(Reader* file, LRFTag* tag)
{
  //<word> size, then string of size bytes
  WORD count;
  if (!file->Read(&count, sizeof(count)))
    return false;
  DWORD size = count+4;
  BYTE* buf = tag->ReserveBuffer(size);
  *(WORD*)buf = count;
  *(WORD*)(buf+count+2)=0;//terminate the string
  return file->Read(buf+2, count);
}

struct TagInfo {
  DWORD paramSize;               //-1 means this tag is invalid
  CustomTagParser pCustomParser; //0 if fixed number of params
};

wchar_t* ObjectTypes[] = { 
  L"<invalid>",//00
  L"PageTree", //01
  L"Page", //02
  L"Header", //03
  L"Footer", //04
  L"PageAtr", //05
  L"Block", //06
  L"BlockAtr", //07
  L"MiniPage", //08
  L"BlockList", //09
  L"Text", //0a
  L"TextAtr", //0b
  L"Image", //0c
  L"Canvas", //0d
  L"ParagraphAtr", //0e
  L"<invalid>",//f0
  L"<invalid>",//10
  L"ImageStream", //11
  L"Import", //12
  L"Button", //13
  L"Window", //14
  L"PopUpWin", //15
  L"Sound", //16
  L"PlaneStream", //17
  L"<invalid>",//18
  L"Font", //19
  L"ObjectInfo", //1a
  L"<invalid>",//1b
  L"BookAtr", //1c
  L"SimpleText", //1d
  L"Toc", //1e
};

std::map<WORD,std::wstring> g_TagNames;

TagInfo StdTags[] = {
  {  6, 0 }, // 00
  {  0, 0 }, // 01
  {  4, 0 }, // 02
  {  4, 0 }, // 03
  {  4, 0 }, // 04
  {  0, 0 }, // 05
  {  0, 0 }, // 06
  {  4, 0 }, // 07
  {  4, 0 }, // 08
  {  4, 0 }, // 09
  {  4, 0 }, // 0A
  {  0, Tag0B_5CParser }, // 0B
  { -1, 0 }, // 0C
  {  0, DummyTagParser }, // 0D
  {  2, 0 }, // 0E
  { -1, 0 }, // 0F
  { -1, 0 }, // 10
  {  2, 0 }, // 11
  {  2, 0 }, // 12
  {  2, 0 }, // 13
  {  2, 0 }, // 14
  {  2, 0 }, // 15
  {  0, TagStringParser }, // 16
  {  4, 0 }, // 17
  {  4, 0 }, // 18
  {  2, 0 }, // 19
  {  2, 0 }, // 1A
  {  2, 0 }, // 1B
  {  2, 0 }, // 1C
  {  2, 0 }, // 1D
  {  2, 0 }, // 1E
  { -1, 0 }, // 1F
  { -1, 0 }, // 20
  {  2, 0 }, // 21
  {  2, 0 }, // 22
  {  2, 0 }, // 23
  {  2, 0 }, // 24
  {  2, 0 }, // 25
  {  2, 0 }, // 26
  {  2, 0 }, // 27
  {  2, 0 }, // 28
  {  6, 0 }, // 29
  {  2, 0 }, // 2A
  {  2, 0 }, // 2B
  {  2, 0 }, // 2C
  {  4, 0 }, // 2D
  {  2, 0 }, // 2E
  { -1, 0 }, // 2F
  { -1, 0 }, // 30
  {  2, 0 }, // 31
  {  2, 0 }, // 32
  {  2, 0 }, // 33
  {  4, 0 }, // 34
  {  2, 0 }, // 35
  {  2, 0 }, // 36
  {  4, 0 }, // 37
  {  2, 0 }, // 38
  {  2, 0 }, // 39
  {  2, 0 }, // 3A
  { -1, 0 }, // 3B
  {  2, 0 }, // 3C
  {  2, 0 }, // 3D
  {  2, 0 }, // 3E
  { -1, 0 }, // 3F
  { -1, 0 }, // 40
  {  2, 0 }, // 41
  {  2, 0 }, // 42
  { -1, 0 }, // 43
  {  4, 0 }, // 44
  {  4, 0 }, // 45
  {  2, 0 }, // 46
  {  2, 0 }, // 47
  {  2, 0 }, // 48
  {  8, 0 }, // 49
  {  8, 0 }, // 4A
  {  4, 0 }, // 4B
  {  4, 0 }, // 4C
  {  0, 0 }, // 4D
  { 12, 0 }, // 4E
  { -1, 0 }, // 4F
  { -1, 0 }, // 50
  {  2, 0 }, // 51
  {  2, 0 }, // 52
  {  4, 0 }, // 53
  {  2, 0 }, // 54
  {  0, TagStringParser }, // 55
  {  0, DummyTagParser }, // 56
  {  2, 0 }, // 57
  {  2, 0 }, // 58
  {  0, TagStringParser }, // 59
  {  0, TagStringParser }, // 5A
  {  4, 0 }, // 5B
  {  0, Tag0B_5CParser }, // 5C
  {  0, TagStringParser }, // 5D
  {  2, 0 }, // 5E
  { -1, 0 }, // 5F
  { -1, 0 }, // 60
  {  2, 0 }, // 61
  {  0, 0 }, // 62
  {  0, 0 }, // 63
  {  0, 0 }, // 64
  {  0, 0 }, // 65
  {  0, 0 }, // 66
  {  0, 0 }, // 67
  {  0, 0 }, // 68
  {  0, 0 }, // 69
  {  0, 0 }, // 6A
  {  0, 0 }, // 6B
  {  8, 0 }, // 6C
  {  0, DummyTagParser }, // 6D
  {  0, 0 }, // 6E
  { -1, 0 }, // 6F
  { -1, 0 }, // 70
  {  0, 0 }, // 71
  {  0, 0 }, // 72
  { 10, 0 }, // 73
  { -1, 0 }, // 74
  {  2, 0 }, // 75
  {  2, 0 }, // 76
  {  2, 0 }, // 77
  {  0, Tag78Parser }, // 78
  {  2, 0 }, // 79
  {  2, 0 }, // 7A
  {  4, 0 }, // 7B
  {  4, 0 }, // 7C
  { -1, 0 }, // 7D
  { -1, 0 }, // 7E
  { -1, 0 }, // 7F
  { -1, 0 }, // 80
  {  0, 0 }, // 81
  {  0, 0 }, // 82
  { -1, 0 }, // 83
  { -1, 0 }, // 84
  { -1, 0 }, // 85
  { -1, 0 }, // 86
  { -1, 0 }, // 87
  { -1, 0 }, // 88
  { -1, 0 }, // 89
  { -1, 0 }, // 8A
  { -1, 0 }, // 8B
  { -1, 0 }, // 8C
  { -1, 0 }, // 8D
  { -1, 0 }, // 8E
  { -1, 0 }, // 8F
  { -1, 0 }, // 90
  { -1, 0 }, // 91
  { -1, 0 }, // 92
  { -1, 0 }, // 93
  { -1, 0 }, // 94
  { -1, 0 }, // 95
  { -1, 0 }, // 96
  { -1, 0 }, // 97
  { -1, 0 }, // 98
  { -1, 0 }, // 99
  { -1, 0 }, // 9A
  { -1, 0 }, // 9B
  { -1, 0 }, // 9C
  { -1, 0 }, // 9D
  { -1, 0 }, // 9E
  { -1, 0 }, // 9F
  { -1, 0 }, // A0
  {  4, 0 }, // A1
  {  0, 0 }, // A2
  { -1, 0 }, // A3
  { -1, 0 }, // A4
  {  0, DummyTagParser }, // A5
  {  0, 0 }, // A6
  {  4, 0 }, // A7
  {  0, 0 }, // A8
  {  0, 0 }, // A9
  {  0, 0 }, // AA
  {  0, 0 }, // AB
  {  0, 0 }, // AC
  {  0, 0 }, // AD
  {  0, 0 }, // AE
  { -1, 0 }, // AF
  { -1, 0 }, // B0
  {  0, 0 }, // B1
  {  0, 0 }, // B2
  {  0, 0 }, // B3
  {  0, 0 }, // B4
  {  0, 0 }, // B5
  {  0, 0 }, // B6
  {  0, 0 }, // B7
  {  0, 0 }, // B8
  {  0, 0 }, // B9
  {  0, 0 }, // BA
  {  0, 0 }, // BB
  {  0, 0 }, // BC
  {  0, 0 }, // BD
  {  0, 0 }, // BE
  { -1, 0 }, // BF
  { -1, 0 }, // C0
  {  0, 0 }, // C1
  {  0, 0 }, // C2
  {  2, 0 }, // C3
  {  0, 0 }, // C4
  {  2, 0 }, // C5
  {  2, 0 }, // C6
  {  0, 0 }, // C7
  {  2, 0 }, // C8
  {  0, 0 }, // C9
  {  2, 0 }, // CA
  {  0, DummyTagParser }, // CB
  {  2, 0 }, // CC
  { -1, 0 }, // CD
  { -1, 0 }, // CE
  { -1, 0 }, // CF
  { -1, 0 }, // D0
  {  0, DummyTagParser }, // D1
  {  0, 0 }, // D2
  { -1, 0 }, // D3
  {  2, 0 }, // D4
  { -1, 0 }, // D5
  {  0, 0 }, // D6
  { 14, 0 }, // D7
  {  4, 0 }, // D8
  {  8, 0 }, // D9
  {  2, 0 }, // DA
  {  2, 0 }, // DB
  {  2, 0 }, // DC
  {  2, 0 }, // DD
  { -1, 0 }, // DE
  { -1, 0 }, // DF
  { -1, 0 }, // E0
  { -1, 0 }, // E1
  { -1, 0 }, // E2
  { -1, 0 }, // E3
  { -1, 0 }, // E4
  { -1, 0 }, // E5
  { -1, 0 }, // E6
  { -1, 0 }, // E7
  { -1, 0 }, // E8
  { -1, 0 }, // E9
  { -1, 0 }, // EA
  { -1, 0 }, // EB
  { -1, 0 }, // EC
  { -1, 0 }, // ED
  { -1, 0 }, // EE
  { -1, 0 }, // EF
  { -1, 0 }, // F0
  {  2, 0 }, // F1
  {  4, 0 }, // F2
  {  4, 0 }, // F3
  {  2, 0 }, // F4
  {  4, 0 }, // F5
  {  4, 0 }, // F6
  {  4, 0 }, // F7
  {  4, 0 }, // F8
  {  6, 0 }, // F9
  { -1, 0 }, // FA
  { -1, 0 }, // FB
  { -1, 0 }, // FC
  { -1, 0 }, // FD
  { -1, 0 }, // FE
  { -1, 0 }, // FF
};
